/*
License text attribution: zlib license at http://www.gzip.org/zlib/zlib_license.html
This license is a modification of the zlib license, changing names and dates.

Copyright (C) 2007 Samson Chee

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

Samson Chee samson@herethere.net
*/

// bubble game.cpp : Defines the entry point for the application.

#include "stdafx.h"

#include <stdio.h>
#include <time.h>
#include <ShellAPI.h>

#include "license.h"

#include "bubble game.h"
#include "bubble_data.h"
#include "bubble_sprite.h"

#define MAX_LOADSTRING 100

#define kWinDimX 320
#define kWinDimY 320

#define kBoxSize 23
#define kBoxBorder 1

#ifndef NDEBUG
#define kTimerDelay 1
#else
#define kTimerDelay 40
#endif

#define kDropSpeed 8
#define kDropAcceleration 2.0f

// Global Variables:
HINSTANCE hInst;								// current instance
TCHAR szTitle[MAX_LOADSTRING];					// The title bar text
TCHAR szWindowClass[MAX_LOADSTRING];			// the main window class name

sBubbleData* theBD;
int theTimerSequence = -1;
int theTimerSubsequence = 0;
bool theAutohint = false;
bool theAutorun = false;

HBITMAP theSprite[4] = {0, 0, 0, 0};
HDC theSpriteDC[4] = {0, 0, 0, 0};
HGDIOBJ theSpriteDCOld[4] = {0, 0, 0, 0};

// Forward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

void flipRect(RECT&);

void handleNew(HWND);
void handleHint(HWND);
void handleHintHelper(HWND);
void handleAutohint(HWND);
void handleAutorun(HWND);
void handlePaint(HWND);
void handleDestroy();
void handleClick(HWND, int, int);
void handleTimer(HWND, bool);

void beginAnimation(HWND);
void finishAnimation(HWND);

void drawBDClearMoving(HDC);
void drawBDClearBorders(HDC);
void drawBDBasic(HDC);
void drawBDSelect(HDC);

int getIndexByColor(tBDColor pColor);
COLORREF getRGBbyIndex(int pIndex);
COLORREF getRGBbyColor(tBDColor pColor);

int APIENTRY _tWinMain(HINSTANCE hInstance,
                       HINSTANCE hPrevInstance,
                       LPTSTR    lpCmdLine,
                       int       nCmdShow)
{
  // init rng
  srand((unsigned int)time(0));
  theBD = bd_init();

  if (!theBD)
  {
    MessageBox(0, "Copyright (C) 2007 - Samson Chee", "Copyright", MB_OK);
  }

  // init sprites
  int theMapSize = kBoxSize - (2 * kBoxBorder);

  float* theBubbleMap = bs_calculate_map(theMapSize, theMapSize);
  for (int i = 0; i < 4; i++)
  {
    theSprite[i] = bs_color_sphere(0, theMapSize, theMapSize, theBubbleMap, getRGBbyIndex(i+1));
    theSpriteDC[i] = CreateCompatibleDC(0);
    theSpriteDCOld[i] = SelectObject(theSpriteDC[i], theSprite[i]);
  }
  bs_free_map(theBubbleMap);

  // TODO: Place code here.
  MSG msg;
  HACCEL hAccelTable;

  // Initialize global strings
  LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
  LoadString(hInstance, IDC_BUBBLEGAME, szWindowClass, MAX_LOADSTRING);
  MyRegisterClass(hInstance);

  // Perform application initialization:
  if (!InitInstance (hInstance, nCmdShow)) 
  {
    return FALSE;
  }

  hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_BUBBLEGAME);

  // Main message loop:
  while (GetMessage(&msg, NULL, 0, 0)) 
  {
    if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  return (int) msg.wParam;
}



//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage are only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
  WNDCLASSEX wcex;

  wcex.cbSize = sizeof(WNDCLASSEX); 

  wcex.style			= CS_HREDRAW | CS_VREDRAW;
  wcex.lpfnWndProc	= (WNDPROC)WndProc;
  wcex.cbClsExtra		= 0;
  wcex.cbWndExtra		= 0;
  wcex.hInstance		= hInstance;
  wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_BUBBLEGAME);
  wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
  //wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
  wcex.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
  wcex.lpszMenuName	= (LPCTSTR)IDC_BUBBLEGAME;
  wcex.lpszClassName	= szWindowClass;
  wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

  return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HANDLE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
  HWND hWnd;

  hInst = hInstance; // Store instance handle in our global variable

  hWnd = CreateWindow(
    szWindowClass,
    szTitle,
    WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
    CW_USEDEFAULT,
    CW_USEDEFAULT,
    kWinDimX,
    kWinDimY,
    NULL,
    NULL,
    hInstance,
    NULL);

  if (!hWnd)
  {
    return FALSE;
  }

  RECT theRect;
  GetClientRect(hWnd, &theRect);
  SetWindowPos(
    hWnd,
    HWND_TOP,
    0, 0,
    kWinDimX + (kWinDimX - theRect.right), kWinDimY + (kWinDimY - theRect.bottom),
    SWP_NOMOVE);

  ShowWindow(hWnd, nCmdShow);
  UpdateWindow(hWnd);

  return TRUE;
}

//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  int wmId, wmEvent;

  switch (message) 
  {
  case WM_COMMAND:
    wmId    = LOWORD(wParam); 
    wmEvent = HIWORD(wParam); 
    // Parse the menu selections:
    switch (wmId)
    {
    case IDM_ABOUT:
      DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
      break;
    case ID_FILE_NEW:
      handleNew(hWnd);
      break;
    case ID_TOOLS_HINT:
      handleHint(hWnd);
      break;
    case ID_TOOLS_AUTOHINT:
      handleAutohint(hWnd);
      break;
    case ID_TOOLS_AUTORUN:
      handleAutorun(hWnd);
      break;
    case IDM_EXIT:
      DestroyWindow(hWnd);
      break;
    default:
      return DefWindowProc(hWnd, message, wParam, lParam);
    }
    break;
  case WM_PAINT:
    handlePaint(hWnd);
    break;
  case WM_DESTROY:
    handleDestroy();
    break;
  case WM_LBUTTONUP:
    handleClick(hWnd, LOWORD(lParam), HIWORD(lParam));
    break;
  case WM_TIMER:
    handleTimer(hWnd, true);
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return 0;
}

// Message handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
  case WM_INITDIALOG:
    {
      SetDlgItemText(hDlg, IDC_EDIT_LICENSE, kLicenseText);
      return TRUE;
    }
    break;

  case WM_COMMAND:
    {
      int wmId, wmEvent;

      wmId    = LOWORD(wParam); 
      wmEvent = HIWORD(wParam); 
      // Parse the menu selections:
      switch (wmId)
      {
      case IDC_BUTTON_GOWEB:
        ShellExecute(NULL, "open", "http://www.herethere.net/bg", NULL, NULL, SW_SHOWNORMAL);;
        break;
      }

  case IDM_ABOUT:
    if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
    {
      EndDialog(hDlg, LOWORD(wParam));
      return TRUE;
    }
    }
    break;
  }
  return FALSE;
}

void flipRect(RECT& pRect)
{
  pRect.bottom = kWinDimY - pRect.bottom;
  pRect.left = kWinDimX - pRect.left;
  pRect.right = kWinDimX - pRect.right;
  pRect.top = kWinDimY - pRect.top;
}

void handleNew(HWND pHWnd)
{
  finishAnimation(pHWnd);

  // re-randomize game data
  bd_new_game(theBD);

  // refresh display
  InvalidateRect(pHWnd, 0, FALSE);
}

void handleHint(HWND pHWnd)
{
  finishAnimation(pHWnd);
  handleHintHelper(pHWnd);
}

void handleHintHelper(HWND pHWnd)
{
  if (theBD->mNumSelected > 1)
  {
    bd_mark_clear(theBD);
    beginAnimation(pHWnd);
  }
  else
  {
    bd_mark_largest_group(theBD);

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void handleAutohint(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, ID_TOOLS_AUTOHINT, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  theAutohint = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_TOOLS_AUTOHINT, theNewMenuState);

  if (theAutohint && !theAutorun)
  {
    finishAnimation(pHWnd);

    if (theBD->mNumSelected <= 1)
    {
      bd_mark_largest_group(theBD);

      // refresh display
      InvalidateRect(pHWnd, 0, FALSE);
    }
  }
}

void handleAutorun(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, ID_TOOLS_AUTORUN, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  theAutorun = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_TOOLS_AUTORUN, theNewMenuState);

  if (theAutorun)
  {
    handleHintHelper(pHWnd);

    theTimerSequence = 6;
    SetTimer(pHWnd, 0, kTimerDelay, 0);
  }
}

void handlePaint(HWND pHWnd)
{
  PAINTSTRUCT thePS;
  HDC theHDC;

  // drawing initialize
  theHDC = BeginPaint(pHWnd, &thePS);
  int theOldRop = GetROP2(theHDC);

  // clear hdc
  drawBDClearMoving(theHDC);

  if (theBD->mNumSelected || (theTimerSequence < 0))
  {
    drawBDClearBorders(theHDC);
  }

  // drawing code
  HPEN thePen = CreatePen(PS_SOLID, kBoxBorder, RGB(0x00,0xff,0xff));
  HGDIOBJ theOldPen = SelectObject(theHDC, thePen);

  MoveToEx(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY), NULL);
  LineTo(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY);
  MoveToEx(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY), NULL);
  LineTo(theHDC, kWinDimX, kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY));

  SelectObject(theHDC, theOldPen);
  DeleteObject(thePen);

  SetTextColor(theHDC, RGB(0xff, 0xff, 0xff));
  SetBkColor(theHDC, RGB(0x00, 0x00, 0x00));

  // remember to pad the ends of the output strings with spaces because it's easier to overwrite existing
  // text that way
  char theBuf[8192];
  float theAverageScore = (theBD->mMoves < 1) ? 0 : (float)((theBD->mScore * 1.0) / (theBD->mMoves * 1.0));

  {
    HFONT theFont = CreateFont(17, 0, 0, 0, 0, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, 0);
    HGDIOBJ theOldFont = SelectObject(theHDC, theFont);

    int theGroupSize = (theTimerSequence < 0) ? theBD->mNumSelected : theBD->mLastGroupSize;
    int theGroupScore = (theTimerSequence < 0) ? groupScore(theBD->mNumSelected) : theBD->mLastGroupScore;

    sprintf(theBuf, "score: %d   current: %d/%d                ", theBD->mScore, theGroupSize, theGroupScore);
    TextOut(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), 4, theBuf, (int)strlen(theBuf));

    SelectObject(theHDC, theOldFont);
    DeleteObject(theFont);
  }

  {
    HFONT theFont = CreateFont(17, 0, 900, 0, 0, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, 0);
    HGDIOBJ theOldFont = SelectObject(theHDC, theFont);

    sprintf(theBuf, "moves: %d   average: %2.2f   largest: %d                ", theBD->mMoves, theAverageScore, theBD->mLargestGroup);
    TextOut(theHDC, 2, kWinDimY - 4, theBuf, (int)strlen(theBuf));

    SelectObject(theHDC, theOldFont);
    DeleteObject(theFont);
  }

  drawBDBasic(theHDC);
  drawBDSelect(theHDC);

  // drawing cleanup
  SetROP2(theHDC, theOldRop);
  EndPaint(pHWnd, &thePS);
}

void handleDestroy()
{
  bd_destroy(theBD);

  for (int i = 0; i < 4; i++)
  {
    if (theSprite[i])
    {
      SelectObject(theSpriteDC[i], theSpriteDCOld[i]);
      DeleteObject(theSprite[i]);
      DeleteDC(theSpriteDC[i]);
    }
  }

  PostQuitMessage(0);
}

void handleClick(HWND pHWnd, int pX, int pY)
{
  finishAnimation(pHWnd);

  int theX = kWinDimX - pX;
  int theY = kWinDimY - pY;

  int theGridX = theX / kBoxSize;
  int theGridY = theY / kBoxSize;

  if ((theGridX >= 0) && (theGridX < kBubbleDataMaxX) && (theGridY >= 0) && (theGridY < kBubbleDataMaxY))
  {
    if ((theBD->mNumSelected > 1) && bd_ismarked(theBD, theGridX, theGridY))
    {
      bd_mark_clear(theBD);
      beginAnimation(pHWnd);
    }
    else
    {
      bd_mark_select(theBD, theGridX, theGridY);
    }
  }
  else
  {
    bd_reset_mark(theBD);
  }

  // refresh display
  InvalidateRect(pHWnd, 0, FALSE);
}

void handleTimer(HWND pHWnd, bool pRefresh)
{
  switch (theTimerSequence)
  {
  case 0 :
    {
      bd_gravityY_pre(theBD);

      theTimerSubsequence = kDropSpeed;
      theTimerSequence++;
    }; break;
  case 1 :
    {
      theTimerSubsequence += (int)(theTimerSubsequence / kDropAcceleration);

      if (theTimerSubsequence >= theBD->mMaxMove * kBoxSize)
      {
        theTimerSequence++;
      }
    }; break;
  case 2 :
    {
      bd_gravityY_post(theBD);

      if (bd_can_load_qx(theBD))
      {
        bd_load_qx(theBD);
        bd_gravityY_pre(theBD);

        theTimerSubsequence = kDropSpeed;
        theTimerSequence = 1;
      }
      else
      {
        bd_gravityX_pre(theBD);

        theTimerSubsequence = kDropSpeed;
        theTimerSequence++;
      }
    }; break;
  case 3 :
    {
      theTimerSubsequence += (int)(theTimerSubsequence / kDropAcceleration);

      if (theTimerSubsequence >= theBD->mMaxMove * kBoxSize)
      {
        theTimerSequence++;
      }
    }; break;
  case 4 :
    {
      bd_gravityX_post(theBD);

      if (bd_can_load_qy(theBD))
      {
        bd_load_qy(theBD);
        bd_gravityX_pre(theBD);

        theTimerSubsequence = kDropSpeed;
        theTimerSequence = 3;
      }
      else
      {
        theTimerSubsequence = 0;
        theTimerSequence++;
      }
    }; break;
  case 5 :
    {
      KillTimer(pHWnd, 0);
      theTimerSequence = -1;

      // check game over
      if (!bd_game_active(theBD))
      {
        // refresh display
        InvalidateRect(pHWnd, 0, FALSE);

        // new game prompt
        if (theAutorun) {
          bd_new_game(theBD);
        } 
        else if (MessageBox(pHWnd, "Game Over - New Game?", "Game Over", MB_YESNO) == IDYES)
        {
          handleNew(pHWnd);
        }
      }

      if (theAutorun)
      {
        handleHintHelper(pHWnd);
        theTimerSequence = 6;
        SetTimer(pHWnd, 0, kTimerDelay, 0);
      }
      else if (theAutohint)
      {
        handleHintHelper(pHWnd);
      }

    }; break;
  case 6 :
    {
      bd_mark_clear(theBD);
      beginAnimation(pHWnd);
    }; break;
  }

  if (pRefresh)
  {
    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void beginAnimation(HWND pHWnd)
{
  theTimerSequence = 0;
  SetTimer(pHWnd, 0, kTimerDelay, 0);
}

void finishAnimation(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_TOOLS_AUTORUN, MF_UNCHECKED);

  theAutorun = false;

  if (theTimerSequence >= 0)
  {
    KillTimer(pHWnd, 0);

    while (theTimerSequence >= 0)
    {
      handleTimer(pHWnd, false);
    }

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void drawBDClearMoving(HDC pHDC)
{
  RECT theRect;
  HBRUSH theBackgroundBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  // matrix
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(theBD, i, j);
      if (theNode->mMark || theNode->mMove || (theNode->mColor == 0))
      {
        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        flipRect(theRect);

        FillRect(pHDC, &theRect, theBackgroundBrush);
      }
    }
  }

}

void drawBDClearBorders(HDC pHDC)
{
  RECT theRect;
  HBRUSH theBackgroundBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  for (tBDDim i = 0; i <= kBubbleDataMaxX; i++)
  {
    theRect.top = 0;
    theRect.bottom = kBubbleDataMaxY * kBoxSize;

    theRect.left = i * kBoxSize;
    theRect.left = (i == 0) ? theRect.left : theRect.left - kBoxBorder;

    theRect.right = theRect.left + kBoxBorder;
    theRect.right = (i == kBubbleDataMaxX) ? theRect.right : theRect.right + kBoxBorder;

    flipRect(theRect);
    FillRect(pHDC, &theRect, theBackgroundBrush);
  }

  for (tBDDim j = 0; j <= kBubbleDataMaxY; j++)
  {
    theRect.top = j * kBoxSize;
    theRect.top = (j == 0) ? theRect.top : theRect.top - kBoxBorder;

    theRect.bottom = theRect.top + kBoxBorder;
    theRect.bottom = (j == kBubbleDataMaxY) ? theRect.bottom : theRect.bottom + kBoxBorder;

    theRect.left = 0;
    theRect.right = kBubbleDataMaxX * kBoxSize;

    flipRect(theRect);
    FillRect(pHDC, &theRect, theBackgroundBrush);
  }

}

void drawBDBasic(HDC pHDC)
{
  RECT theRect;

  // matrix
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(theBD, i, j);
      if (theNode->mColor)
      {
        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        if (theNode->mMove)
        {
          switch (theTimerSequence)
          {
          case 1:
          case 2:
            {
              if (theTimerSubsequence < theNode->mMove * kBoxSize)
              {
                theRect.top -= theTimerSubsequence;
                theRect.bottom -= theTimerSubsequence;
              }
              else
              {
                theRect.top -= theNode->mMove * kBoxSize;
                theRect.bottom -= theNode->mMove * kBoxSize;
              }
            }; break;
          case 3:
          case 4:
            {
              if (theTimerSubsequence < theNode->mMove * kBoxSize)
              {
                theRect.left -= theTimerSubsequence;
                theRect.right -= theTimerSubsequence;
              }
              else
              {
                theRect.left-= theNode->mMove * kBoxSize;
                theRect.right -= theNode->mMove * kBoxSize;
              }
            }; break;
          }
        }

        InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
        flipRect(theRect);
        BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
      }
    }
  }

  // draw queues

  // x queue 
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bd_ptr_qx(theBD, i);

    theRect.top = kBubbleDataMaxY * kBoxSize;
    theRect.bottom = (kBubbleDataMaxY + 1) * kBoxSize;
    theRect.left = i * kBoxSize;
    theRect.right = (i + 1) * kBoxSize;

    InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
    flipRect(theRect);
    BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
  }

  // y queue
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    sBubbleNode* theNode = bd_ptr_qy(theBD, j);

    theRect.top = j * kBoxSize;
    theRect.bottom = (j + 1) * kBoxSize;
    theRect.left = kBubbleDataMaxX * kBoxSize;
    theRect.right = (kBubbleDataMaxX + 1) * kBoxSize;

    InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
    flipRect(theRect);
    BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
  }

}

void drawBDSelect(HDC pHDC)
{
  if (theBD->mNumSelected <= 1)
  {
    return;
  }

  HBRUSH thePosBrush = (HBRUSH)GetStockObject(WHITE_BRUSH);
  HBRUSH theNegBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(theBD, i, j);
      if (theNode->mMark)
      {
        RECT theRect;

        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        flipRect(theRect);

        //FillRect(pHDC, &theRect, thePosBrush);

        bool theDrawFollowing = false;

        // side cases
        theDrawFollowing = true;
        if (i > 0)
        {
          theDrawFollowing = !(bd_ptr(theBD, i - 1, j)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (i < kBubbleDataMaxX - 1)
        {
          theDrawFollowing = !(bd_ptr(theBD, i + 1, j)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (j > 0)
        {
          theDrawFollowing = !(bd_ptr(theBD, i, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (j < kBubbleDataMaxY - 1)
        {
          theDrawFollowing = !(bd_ptr(theBD, i, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        // corner cases

        theDrawFollowing = true;
        if ((i > 0) && (j > 0))
        {
          theDrawFollowing = !(bd_ptr(theBD, i - 1, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i > 0) && (j < kBubbleDataMaxY - 1))
        {
          theDrawFollowing = !(bd_ptr(theBD, i - 1, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i < kBubbleDataMaxX - 1) && (j > 0))
        {
          theDrawFollowing = !(bd_ptr(theBD, i + 1, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i < kBubbleDataMaxX - 1) && (j < kBubbleDataMaxY - 1))
        {
          theDrawFollowing = !(bd_ptr(theBD, i + 1, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

      }
    }
  }
}

int getIndexByColor(tBDColor pColor)
{
  switch (pColor)
  {
  case 0x00 : return 255;
  case 0xff : return 254;
  case 0x01 : return 1;
  case 0x02 : return 2;
  case 0x04 : return 3;
  case 0x08 : return 4;
  default : return 0;
  }
}

COLORREF getRGBbyIndex(int pIndex)
{
  switch (pIndex)
  {
  case 255 : return RGB(0xff,0xff,0xff);
  case 254 : return RGB(0x7f,0x7f,0x7f);
  case   1 : return RGB(0xff,0x00,0x00);
  case   2 : return RGB(0x00,0xff,0x00);
  case   3 : return RGB(0x60,0x60,0xff);
  case   4 : return RGB(0xff,0xd7,0x00);
  default : return RGB(0x00,0x00,0x00);
  }
}

COLORREF getRGBbyColor(tBDColor pColor)
{
  return getRGBbyIndex(getIndexByColor(pColor));
}
