/*
License text attribution: zlib license at http://www.gzip.org/zlib/zlib_license.html
This license is a modification of the zlib license, changing names and dates.

Copyright (C) 2007 Samson Chee

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

Samson Chee samson@herethere.net
*/

#include "stdafx.h"

#include <stdio.h>
#include "bubble_data.h"

tBDBool isOneColor(tBDColor pColor)
{
  return
    (pColor == (1 << 0)) ||
    (pColor == (1 << 1)) ||
    (pColor == (1 << 2)) ||
    (pColor == (1 << 3)) ||
    (pColor == (1 << 4)) ||
    (pColor == (1 << 5)) ||
    (pColor == (1 << 6)) ||
    (pColor == (1 << 7));
}

tBDColor getRandColor()
{
  return (1 << (rand() % 4));
}

int groupScore(tBDDim2 pSize)
{
  int theAcc = 0;

  if (pSize == 0)
  {
    return 0;
  }

  for (int i = 2; i <= pSize; i++)
  {
    theAcc += i;
  }

  return theAcc;
}

sBubbleData* bd_init()
{
  // init 2d matrix and also both 1d queue matrixes
  tBDDim2 theNumElements = (kBubbleDataMaxX * kBubbleDataMaxY) + kBubbleDataMaxX + kBubbleDataMaxY;
  sBubbleData* theData = (sBubbleData*)malloc(sizeof(sBubbleData) + (theNumElements * sizeof(sBubbleNode)));

  bd_new_game(theData);

  return theData;
}

void bd_destroy(sBubbleData* pBD)
{
  free(pBD);
}

tBDBool bd_game_active(sBubbleData* pBD)
{
  // return true if there is at least one pair (which means that the game is not over yet)
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      tBDBool theOk = (bd_mark_select(pBD, i, j) > 1);
      bd_reset_mark(pBD);
      bd_reset_hit(pBD);

      if (theOk)
      {        
        return 1;
      }
    }
  }

  return 0;
}

void bd_new_game(sBubbleData* pBD)
{
  bd_reset(pBD);

  bd_randomize_color(pBD);
  while (!bd_game_active(pBD))
  {
    bd_randomize_color(pBD);
  }

  bd_randomize_color_qx(pBD);
  bd_randomize_color_qy(pBD);

  pBD->mNumSelected = 0;
  pBD->mLargestGroup = 0;
  pBD->mMaxMove = 0;
  pBD->mMoves = 0;
  pBD->mScore = 0;

  pBD->mLastGroupSize = 0;
  pBD->mLastGroupScore = 0;
}

sBubbleNode* bd_ptr_1d(sBubbleData* pBD, tBDDim2 pNum)
{
  if (pNum >= kBubbleDataMaxX * kBubbleDataMaxY)
  {
    return 0;
  }

  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + pNum;
}

sBubbleNode* bd_ptr(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  if ((pX >= kBubbleDataMaxX) || (pY >= kBubbleDataMaxY))
  {
    return 0;
  }

  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + ((pY * kBubbleDataMaxX) + pX);
}

void bd_reset(sBubbleData* pBD)
{
  memset(pBD, 0, kBubbleDataMaxX * kBubbleDataMaxY * sizeof(sBubbleNode));
}

void bd_reset_hit(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mHit = 0;
  }
}

void bd_randomize_color(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mColor = getRandColor();
  }
}

void bd_reset_mark(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mMark = 0;
  }

  pBD->mNumSelected = 0;
}

void bd_reset_move(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mMove = 0;
  }
}

void bd_mark(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  bd_ptr(pBD, pX, pY)->mMark = 1;
}

tBDBool bd_ismarked(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  return bd_ptr(pBD, pX, pY)->mMark;
}

tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  bd_reset_mark(pBD);
  pBD->mNumSelected = bd_mark_select(pBD, pX, pY, 0xff);

  return pBD->mNumSelected;
}

tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY, tBDColor pColor)
{
  sBubbleNode* theNode = bd_ptr(pBD, pX, pY);

  // ignore marked or hit nodes because we've already been here
  if ((theNode->mColor == 0) || (theNode->mHit & theNode->mColor) || (theNode->mMark) || (theNode->mMove))
  {
    return 0;
  }

  // node is not marked, but it may already have been hit by another color
  // - as long as we remember to check the bit flags, we'll be fine

  // hit the node if the colors don't match
  if ((theNode->mColor & pColor) == 0)
  {
    theNode->mHit |= pColor;
    return 0;
  }

  // mark this node
  theNode->mMark = 1;

  // go look at neighbouring nodes
  tBDDim2 theAcc = 1;
  tBDColor theNodeColor = isOneColor(theNode->mColor) ? theNode->mColor : pColor;

  if (pX > 0)
  {
    theAcc += bd_mark_select(pBD, pX - 1, pY, theNodeColor);
  }

  if (pX < kBubbleDataMaxX - 1)
  {
    theAcc += bd_mark_select(pBD, pX + 1, pY, theNodeColor);
  }

  if (pY > 0)
  {
    theAcc += bd_mark_select(pBD, pX, pY - 1, theNodeColor);
  }

  if (pY < kBubbleDataMaxY - 1)
  {
    theAcc += bd_mark_select(pBD, pX, pY + 1, theNodeColor);
  }

  return theAcc;
}

tBDDim2 bd_mark_clear(sBubbleData* pBD)
{
  tBDDim2 theAcc = 0;

  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    sBubbleNode* theBD = bd_ptr_1d(pBD, i);

    // clear only the marked nodes
    if (theBD->mMark)
    {
      theBD->mColor = 0;
      theAcc++;
    }

    theBD->mHit = 0;
    theBD->mMark = 0;
    theBD->mMove = 0;
  }

  pBD->mNumSelected = 0;
  pBD->mLargestGroup = (pBD->mLargestGroup > theAcc) ? pBD->mLargestGroup : theAcc;
  pBD->mMoves++;

  pBD->mLastGroupSize = theAcc;
  pBD->mLastGroupScore = groupScore(theAcc);

  pBD->mScore += pBD->mLastGroupScore;

  return theAcc;
}

tBDDim2 bd_mark_largest_group(sBubbleData* pBD)
{
  // get the largest group by looking at each node
  // - mark each group that we finish looking at using the mMove member variable as temporary storage
  // - groups of 1 are always marked as 1
  // - groups smaller than the largest group are also marked as 1

  tBDDim2 theLargestGroup = 0;

  tBDDim i;
  tBDDim j;

  for (i = 0; i < kBubbleDataMaxX; i++)
  {
    for (j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if ((theNode->mColor == 0) || (theNode->mMove))
      {
        continue;
      }

      tBDDim2 theGroupSize = bd_mark_select(pBD, i, j);

      for (tBDDim x = 0; x < kBubbleDataMaxX; x++)
      {
        for (tBDDim y = 0; y < kBubbleDataMaxY; y++)
        {
          sBubbleNode* theNode = bd_ptr(pBD, x, y);

          if (theNode->mColor == 0)
          {
            continue;
          }

          if (theGroupSize > theLargestGroup)
          {
            if (theNode->mMark)
            {
              theNode->mMove = 2;
            }
            else
            {
              if (theNode->mMove > 1)
              {
                theNode->mMove = 1;
              }
            }
          }
          else
          {
            theNode->mMove = (theNode->mMark) ? 1 : theNode->mMove;
          }            
        }
      }

      theLargestGroup = (theGroupSize > theLargestGroup) ? theGroupSize : theLargestGroup;
    }
  }

  pBD->mNumSelected = 0;

  for (i = 0; i < kBubbleDataMaxX; i++)
  {
    for (j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if (theNode->mMove > 1)
      {
        theNode->mMark = 1;
        pBD->mNumSelected++;
      }
      else
      {
        theNode->mMark = 0;
      }

      theNode->mMove = 0;
    }
  }

  return pBD->mNumSelected;
}

void bd_move_node(sBubbleData* pBD, tBDDim pOrgX, tBDDim pOrgY, tBDDim pNewX, tBDDim pNewY)
{
  sBubbleNode* theOrgNode = bd_ptr(pBD, pOrgX, pOrgY);
  sBubbleNode* theNewNode = bd_ptr(pBD, pNewX, pNewY);

  theNewNode->mColor = theOrgNode->mColor;
  theOrgNode->mColor = 0;

  theNewNode->mMove = 0;
  theOrgNode->mMove = 0;
}

void bd_gravityX_pre(sBubbleData* pBD)
{
  // calculate how much to drop
  tBDDim theMoveAcc = 0;
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    tBDBool theColEmpty = 1;
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mColor)
      {
        theNode->mMove = theMoveAcc;
        theColEmpty = 0;

        pBD->mMaxMove = (pBD->mMaxMove > theMoveAcc) ? pBD->mMaxMove : theMoveAcc;
      }
    }

    if (theColEmpty)
    {
      theMoveAcc++;
    }
  }
}

void bd_gravityX_post(sBubbleData* pBD)
{
  // drop everything down
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mMove)
      {
        // move to last hole and update last hole to the position above
        bd_move_node(pBD, i, j, i - theNode->mMove, j);
      }
    }
  }

  pBD->mMaxMove = 0;
}

void bd_gravityY_pre(sBubbleData* pBD)
{
  // calculate how much to drop
  tBDDim theMoveAcc = 0;
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    theMoveAcc = 0;
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mColor)
      {
        theNode->mMove = theMoveAcc;
        pBD->mMaxMove = (pBD->mMaxMove > theMoveAcc) ? pBD->mMaxMove : theMoveAcc;
      }
      else
      {
        theMoveAcc++;
      }
    }
  }
}

void bd_gravityY_post(sBubbleData* pBD)
{
  // drop everything down
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mMove)
      {
        // move to last hole and update last hole to the position above
        bd_move_node(pBD, i, j, i, j - theNode->mMove);
      }
    }
  }

  pBD->mMaxMove = 0;
}

sBubbleNode* bd_ptr_qx(sBubbleData* pBD, tBDDim pNum)
{
  if (pNum >= kBubbleDataMaxX)
  {
    return 0;
  }

  tBDDim2 theOffset = (kBubbleDataMaxX * kBubbleDataMaxY) + pNum;
  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + theOffset;
}

sBubbleNode* bd_ptr_qy(sBubbleData* pBD, tBDDim pNum)
{
  if (pNum >= kBubbleDataMaxY)
  {
    return 0;
  }

  tBDDim2 theOffset = (kBubbleDataMaxX * kBubbleDataMaxY) + kBubbleDataMaxX + pNum;
  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + theOffset;
}

void bd_randomize_color_qx(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bd_ptr_qx(pBD, i);
    theNode->mColor = getRandColor();
  }
}

void bd_randomize_color_qy(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theNode = bd_ptr_qy(pBD, i);
    theNode->mColor = getRandColor();
  }
}

tBDBool bd_can_load_qx(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bd_ptr(pBD, i, kBubbleDataMaxY - 1);
    if (theNode->mColor)
    {
      return 0;
    }
  }

  return 1;
}

tBDBool bd_can_load_qy(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theNode = bd_ptr(pBD, kBubbleDataMaxX - 1, i);
    if (theNode->mColor)
    {
      return 0;
    }
  }

  return 1;
}

void bd_load_qx(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theMatrixNode = bd_ptr(pBD, i, kBubbleDataMaxY - 1);
    sBubbleNode* theQueueNode = bd_ptr_qx(pBD, i);

    theMatrixNode->mColor = theQueueNode->mColor;
  }

  bd_randomize_color_qx(pBD);
}

void bd_load_qy(sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theMatrixNode = bd_ptr(pBD, kBubbleDataMaxX - 1, i);
    sBubbleNode* theQueueNode = bd_ptr_qy(pBD, i);

    theMatrixNode->mColor = theQueueNode->mColor;
  }

  bd_randomize_color_qy(pBD);
}

void bd_debug_output_matrix(sBubbleData* pBD)
{
#ifdef WIN32
  // print color values
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      char theBuf[4];
      sprintf(theBuf, "%02X ", theNode->mColor);
      OutputDebugString(theBuf);
    }

    OutputDebugString("\n");
  }
#endif
}
