/*
  License text attribution: zlib license at http://www.gzip.org/zlib/zlib_license.html
  This license is a modification of the zlib license, changing names and dates.

  Copyright (C) 2007 Samson Chee

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Samson Chee samson@herethere.net
*/

#ifndef __bubble_data_h__
#define __bubble_data_h__

#define kBubbleDataMaxX 12
#define kBubbleDataMaxY 12

// data types
typedef unsigned char tBDDim;
typedef unsigned short tBDDim2;
typedef unsigned char tBDColor;
typedef unsigned char tBDBool;

// basic data

// node
// make sure that this is 32-bit aligned, i.e. sizeof(structure) is a multiple of 32 bits
struct sBubbleNode
{
  tBDColor mHit;   // bit map: hit by color
  tBDColor mColor; // bit map: color
  tBDBool  mMark;  // boolean
  tBDDim   mMove;  // move direction dependent on context
};

// container
struct sBubbleData
{
  tBDDim2 mNumSelected;
  tBDDim2 mLargestGroup;

  tBDDim mMaxMove;

  int mMoves;
  int mScore;

  int mLastGroupSize;
  int mLastGroupScore;

  sBubbleNode mNodes[1];
};

// functions

tBDBool isOneColor(tBDColor pColor);
tBDColor getRandColor();
int groupScore(tBDDim2);

sBubbleData* bd_init();
void bd_destroy(sBubbleData* pBD);

tBDBool bd_game_active(sBubbleData* pBD);
void bd_new_game(sBubbleData* pBD);

sBubbleNode* bd_ptr_1d(sBubbleData* pBD, tBDDim2 pNum);
sBubbleNode* bd_ptr(sBubbleData* pBD, tBDDim pX, tBDDim pY);

void bd_reset(sBubbleData* pBD);
void bd_reset_hit(sBubbleData* pBD);
void bd_randomize_color(sBubbleData* pBD);
void bd_reset_mark(sBubbleData* pBD);
void bd_reset_move(sBubbleData* pBD);

void bd_mark(sBubbleData* pBD, tBDDim pX, tBDDim pY);
tBDBool bd_ismarked(sBubbleData* pBD, tBDDim pX, tBDDim pY);

tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY);
tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY, tBDColor pColor);
tBDDim2 bd_mark_clear(sBubbleData* pBD);

tBDDim2 bd_mark_largest_group(sBubbleData* pBD);

void bd_move_node(sBubbleData* pBD, tBDDim pOrgX, tBDDim pOrgY, tBDDim pNewX, tBDDim pNewY);

void bd_gravityX_pre(sBubbleData* pBD);
void bd_gravityX_post(sBubbleData* pBD);
void bd_gravityY_pre(sBubbleData* pBD);
void bd_gravityY_post(sBubbleData* pBD);

sBubbleNode* bd_ptr_qx(sBubbleData* pBD, tBDDim pNum);
sBubbleNode* bd_ptr_qy(sBubbleData* pBD, tBDDim pNum);

void bd_randomize_color_qx(sBubbleData* pBD);
void bd_randomize_color_qy(sBubbleData* pBD);

tBDBool bd_can_load_qx(sBubbleData* pBD);
tBDBool bd_can_load_qy(sBubbleData* pBD);

void bd_load_qx(sBubbleData* pBD);
void bd_load_qy(sBubbleData* pBD);

void bd_debug_output_matrix(sBubbleData* pBD);

#endif // __bubble_data_h__
